/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* discover.c

Here is a rough outline of the discovery procedure:

   Given: S = {substructures to be tried}
          L = limit on computation time

   Do:    D = {}
          order S from best to worst
          while (computation < L) and (S <> {}) do
            b = (first S)
            insert b into D, best first, no duplicates
            B = b extended in all possible ways
            insert each sub in B into S, best first, no duplicates
          return D
*/

#include "subdue.h"


sub_list_node_ptr discover_subs(subs)
     sub_list_node_ptr subs;
/* Returns a list of (<=NumSubs) discovered substructures in order from best
   to worst.  Discover_subs assumes that the substructures on the input list
   subs are all unique. */
{
  shortint i;
  sub_list_node_ptr dsubs = NULL;   /* list of discovered substructures */
  sub_list_node_ptr sublist = NULL; /* sorted list of substructures */
  sub_list_node_ptr sptr;
  substructure *s;
  shortint computation = 0;
  sub_list_node_ptr temp, temp1;
  graph_list_node_ptr gtemp;
  float value;

  sptr = subs;               /* put given subs on sublist */
  while (sptr != NULL) {
    sublist = insert_sub(sptr->s,sublist,FALSE); /* sublist becomes sorted */
    temp = sptr;
    sptr = sptr->next;
    free(temp);
  }
  if (Trace > 0) {
    printf("\nInitial substructures:\n\n");
    print_sub_list(sublist);
  }
  while ((sublist != NULL) && (computation < Limit)) {
    computation++;
    sptr = sublist;                     /* remove sub from sublist */
    sublist = sublist->next;
    s = sptr->s;
    free(sptr);
    if (Trace > 0) {
      printf("\nConsidering substructure:\n\n");
      print_substructure(s);
    }
    if (computation < Limit) {
      value = s->value;
      sptr = extend_sub(s);     /* add extensions to sublist */
      while (sptr != NULL) {
	evaluate_sub(sptr->s);
	if ((! Prune) || (Prune && (sptr->s->value >= value)))
	  if (! member_sub(sptr->s->definition, dsubs))
	    sublist = insert_sub(sptr->s,sublist,TRUE);
	temp = sptr;
	sptr = sptr->next;
	free(temp);
      }
      i = 1;
      temp = sublist; /* limit substructure queue to length Beam */
      while ((temp != NULL) && (i < Beam)) {
	temp = temp->next;
	i++;
      }
      if (temp != NULL) {  /* free extra subs on sublist */
	temp1 = temp->next;
	temp->next = NULL;
	while (temp1 != NULL) {
	  temp = temp1;
	  temp1 = temp1->next;
	  free_substructure(temp->s);
	  free(temp);
	}
      }
    }
    dsubs = best_n_subs(s,dsubs,NumSubs);
  }
  return dsubs;  
}


sub_list_node_ptr best_n_subs(sub,dlist,n)
     substructure *sub;
     sub_list_node_ptr dlist;
     shortint n;
/* Adds substructure sub to proper place in dlist ordered by increasing
   substructure value, but contrains dlist to contain only the best n
   substructures.  New pointer to dlist returned. */
{
  shortint i = 0;
  sub_list_node_ptr dptr, prev = NULL;
  substructure *s, *stmp;

  s = sub;
  if (dlist == NULL) {
    dlist = (sub_list_node_ptr) malloc(sizeof(substructure_list_node));
    dlist->s = s;
    dlist->next = NULL;
  } else {
    dptr = dlist;
    while ((i < n) && (dptr != NULL)) { /* put s in proper place */
      if (s->value > dptr->s->value) {
	stmp = dptr->s;
	dptr->s = s;
	s = stmp;
      }
      i++;
      prev = dptr;
      dptr = dptr->next;
    }
    if (i < n) {  /* room for more substructures on dlist */
      dptr = (sub_list_node_ptr) malloc(sizeof(substructure_list_node));
      dptr->s = s;
      dptr->next = NULL;
      prev->next = dptr;
    } else free_substructure(s);
  }
  return(dlist);
}


sub_list_node_ptr insert_sub(s, slist, new)
     substructure *s;
     sub_list_node_ptr slist;
     boolean new;
/* Destructively inserts substructure s into slist in order of increasing
   substructure value.  If new=TRUE, then only substructures not already
   on slist are inserted.  The new list is returned. */
{
  sub_list_node_ptr slnptr, slnptr1;
  boolean done = FALSE;

  if ((! new) || (! member_sub(s->definition,slist))) {
    slnptr = (sub_list_node_ptr) malloc(sizeof(substructure_list_node));
    slnptr->s = s;
    slnptr->next = NULL;
    if (slist == NULL)
      slist = slnptr;
    else if (s->value > slist->s->value) {
      slnptr->next = slist;
      slist = slnptr;
    } else {
      slnptr1 = slist;
      while ((! done) && (slnptr1->next != NULL))
	if (s->value > slnptr1->next->s->value)
	  done = TRUE;
	else slnptr1 = slnptr1->next;
      slnptr->next = slnptr1->next;
      slnptr1->next = slnptr;
    }
  }
  return slist;
}


boolean member_sub(g, slist)
     graphptr g;
     sub_list_node_ptr slist;
/* Returns TRUE if the given graph exactly matches a substructure on slist. */
{
  boolean match = FALSE;
  while ((slist != NULL) && (! match))
    if (exact_gmatch(g,slist->s->definition))
      match = TRUE;
    else slist = slist->next;
  return match;
}

