/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* eval.c

The evaluation of a substructure consists of determining its
description length and then modifying the inverse of this value based
on the presence or absence of domain specific and independent
properties or rules.  Currently, three such rules are implemented:

      connectivity - the inverse of the average connectivity per
instance of the substructure, where the connectivity of an instance is
the number of external connections for that instance

      compactness - the ratio of the number of edges in the
substructure to the number of vertices in the substructure

      coverage - the ratio of the amount of the graph covered by the
instances of the substructure to the size of the graph

These values (which must be greater than or equal to one) are then
raised to the power Connectivity, Compactness, and Coverage
respectively.  A positive power favors the property, a negative power
favors the absence of the property, and a power of 0 ignores the
property.  The product of the description length and these property
values determines the value of a substructure.

TO ADD MORE RULES (i.e., properties, heuristics):
  1. Write code accepting a (substructure *) as input and returning
     a (float) >= 1, where a higher value indicates the property.
  2. Add an argument to the subdue.c main procedure so that the user
     may supply an exponent (define variable in subdue.h).
  3. Add a line to the 'rules' function to include your new rule.

*/

#include "subdue.h"


void evaluate_sub(s)
     substructure *s;
/* Sets the value of substructure s.  First, the value consists of
   the inverse of the description length of the substructure plus the
   description length of the original Graph compressed by the substructure.
   Then, this value is modified by multiplying the result of evaluating the
   rules. */
{
  shortint i;
  float value;
  graphptr ng;
  
  value = (float) mdl(s->definition);
  ng = compress(Graph, s, 0);  /* create compressed graph */
  value += (float) mdl(ng);
  value = 1.0 / value;
  for (i=0; i<ng->nv; i++)     /* free compressed graph */
    free(ng->vertices[i]);
  for (i=0; i<ng->ne; i++)
    free(ng->edges[i]);
  free(ng->vertices);
  free(ng->edges);
  free(ng);
  s->value = value * rules(s);
}


float rules(s)
     substructure *s;
/* Returns the product of the value of each rule raised to its globally-
   defined exponent. */
{
  float value = 1.0;

  if (Connectivity != 0.0)
    value *= (float) pow((double) connectivity(s), (double) Connectivity);
  if (Compactness != 0.0)
    value *= (float) pow((double) compactness(s), (double) Compactness);
  if (Coverage != 0.0)
    value *= (float) pow((double) coverage(s), (double) Coverage);
  return(value);
}


float unique_structure(s)
     substructure *s;
/*                  ---
                    \
 unique_structure = /    weight(i) * unique_structure(i)
                    ---
                   i in I

  where I is the set of instances of substructure s, weight(i) is [1 -
  (matchcost(s,i)/size(i))], and unique_structure(i) is the amount of
  of structure in the original graph G (not counted in previous terms
  of the summation) covered by instance i.
*/
{
  vptr *varray; /* vertices seen so far */
  shortint vindex = 0;
  vptr v;
  eptr *earray; /* edges seen so far */
  shortint eindex = 0;
  eptr e;
  shortint uv, ue; /* unique vertices and edges */
  shortint i;
  float total = 0.0;
  graph_list_node_ptr glnptr = s->instances;

  varray = (vptr *) malloc(Graph->nv * sizeof(vptr));
  earray = (eptr *) malloc(Graph->ne * sizeof(eptr));
  while (glnptr != NULL) {
    uv = 0; ue = 0;
    for (i = 0; i < glnptr->g->nv; i++) {
      v = glnptr->g->vertices[i];
      if (! member_vertex(v,varray,vindex)) {
	uv++;
	varray[vindex] = v;
	vindex++;
      }
    }
    for (i = 0; i < glnptr->g->ne; i++) {
      e = glnptr->g->edges[i];
      if (! member_edge(e,earray,eindex)) {
	ue++; earray[eindex] = e; eindex++;}
    }
    total = total + (instance_weight(glnptr->g) * (uv + ue));
    glnptr = glnptr->next;
  }
  free(varray);
  free(earray);
  return total;
}


boolean member_vertex(v, varray, vindex)
     vptr v, *varray;
     shortint vindex;
/* Returns TRUE if vertex v is in varray. */
{
  boolean member = FALSE;
  shortint i = 0;
  while ((i < vindex) && (! member))
    if (varray[i] == v) member = TRUE; else i++;
  return member;
}


boolean member_edge(e, earray, eindex)
     eptr e, *earray;
     shortint eindex;
/* Returns TRUE if edge e is in earray. */
{
  boolean member = FALSE;
  shortint i = 0;
  while ((i < eindex) && (! member))
    if (earray[i] == e) member = TRUE; else i++;
  return member;
}


float instance_weight(g)
     graphptr g;
{
  float size = (float) (g->nv + g->ne);
  if (g->matchcost > size)
    size = g->matchcost;
  return (1.0 - (g->matchcost / size));
}


float connectivity(s)
     substructure *s;
/*                         ---                                     -1
                    [  1   \                                      ]
 connectivity = 1 + [ ---  /    weight(i) * num_external_conns(i) ]
                    [ |I|  ---                                    ]
                          i in I

  where I is the set of instances of substructure s, weight(i) is [1 -
  (matchcost(s,i)/size(i))], and num_external_conns(i) is the number of
  edges connecting a vertex in the instance to a vertex outside the
  instance.
*/
{
  shortint ninsts = 0;
  float total = 0.0;
  shortint i, connections;
  vptr v1, v2;
  eptr e;
  boolean v1p, v2p;
  graph_list_node_ptr glnptr = s->instances;

  while (glnptr != NULL) {
    ninsts++; connections = 0;
    for (i = 0; i < Graph->ne; i++) {
      e = Graph->edges[i];
      v1 = find_vertex(e->from, Graph);
      v2 = find_vertex(e->to, Graph);
      v1p = member_vertex(v1,glnptr->g->vertices,glnptr->g->nv);
      v2p = member_vertex(v2,glnptr->g->vertices,glnptr->g->nv);
      if ((v1p && (! v2p)) || (v2p && (! v1p))) connections++;
    }
    total = total + (instance_weight(glnptr->g) * connections);
    glnptr = glnptr->next;
  }
  if (total == 0)  /* Catch-all - this will make no connections have C = 1 */
    total = ninsts;
  return (1.0 + ((float) ninsts / total));
}


float compactness(s)
     substructure *s;
/*                       ---
                     1   \                 edges(i)
 compactness = 1 +  ---  /    weight(i) * -----------
                    |I|  ---              vertices(i)
                        i in I

  where I is the set of instances of substructure s, and weight(i)
  is [1 - (matchcost(s,i)/size(i))].
*/
{
  float total = 0.0;
  shortint ninsts = 0;
  float tmp;
  graph_list_node_ptr glnptr = s->instances;
  
  while (glnptr != NULL) {
    ninsts++;
    tmp = (float) glnptr->g->ne / (float) glnptr->g->nv;
    total = total + (instance_weight(glnptr->g) * tmp);
    glnptr = glnptr->next;
  }
  if (ninsts == 0)
    return(1.0);
  else return(1.0 + (total / (float) ninsts));
}


float coverage(s)
     substructure *s;
/*                        ---
                    1     \               
 coverage = 1 +  -------  /    weight(i) * unique_structure(i)
                 size(G)  ---             
                         i in I

  where I is the set of instances of substructure s, weight(i) is [1 -
  (matchcost(s,i)/size(i))], and unique_structure(i) is the amount of
  of structure in the original graph G (not counted in previous terms
  of the summation) covered by instance i.
*/
{
  float tmp = (unique_structure(s) / (float) Size);
  return (1.0 + tmp);
}
