/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* extend.c

Rough outline of substructure extension process:

   Given: substucture s
          newsubs = {}

   Do:    new-instances = add_edge(instances(s))
          foreach i in new-instances do
            if (not exact_member(i,newsubs)) then
               newsub = make_substructure
               definition(newsub) = i
               foreach j in (new-instances + instances(s)) do
                 if (gmatch(j,newsub) <= Threshold) then
                    instances(newsub) = instances(newsub) + j
               newsubs = newsubs + newsub
          return newsubs
*/

#include "subdue.h"


sub_list_node_ptr extend_sub(s)
     substructure *s;
/* Returns list of substructures representing all possible
   ways of extending the given substructure by one edge. */
{
  sub_list_node_ptr slnptr, newsubs = NULL;
  substructure *newsub;
  graph_list_node_ptr glnptr, newinsts;

  if (Trace > 2) {
    printf("\nEntering extend_sub with substructure:\n");
    print_substructure(s); printf("\n");
  }
  newinsts = add_edge(s->instances);
  glnptr = newinsts;
  while (glnptr != NULL) {
    if (! member_sub(glnptr->g, newsubs)) {
      newsub = (substructure *) malloc(sizeof(substructure));
      glnptr->g->refcount++;
      newsub->definition = glnptr->g;
      newsub->definition->matchcost = 0.0;
      newsub->instances = NULL;
      add_instances(s->instances, newsub);
      add_instances(newinsts, newsub);
      slnptr = (sub_list_node_ptr) malloc(sizeof(substructure_list_node));
      slnptr->s = newsub;
      slnptr->next = newsubs;
      newsubs = slnptr;
    }
    glnptr = glnptr->next;
  }
  while (newinsts != NULL) {  /* Delete all of the new instances */
    glnptr = newinsts;
    newinsts = newinsts->next;
    if (glnptr->g->refcount == 0)
      free(glnptr->g);
    free(glnptr);
  }
  if (Trace > 2) {
    printf("\nExiting extend_sub with substructures:\n");
    if (newsubs != NULL) {
      print_sub_list(newsubs);
      printf("\n");
    }
  }
  return newsubs;
}


void add_instances(insts, sub)
     graph_list_node_ptr insts;
     substructure *sub;
/* Add instances insts to the instances of substructure sub, if they do
   not already exist as instances of sub. */
{
  graphptr ig, newg;
  graph_list_node_ptr glnptr;

  while (insts != NULL) {
    ig = insts->g;
    newg = build_graph(ig->nv, ig->ne, ig->vertices, ig->edges, 0.0);
    newg->matchcost = gmatch(newg, sub->definition);
    if ((newg->matchcost <= ((float) size(newg) * Threshold)) &&
	unique_instance(newg, sub)) {
      glnptr = (graph_list_node_ptr) malloc(sizeof(graph_list_node));
      glnptr->g = newg;
      glnptr->next = sub->instances;
      sub->instances = glnptr;
    } else free(newg);
    insts = insts->next;
  }
}


boolean unique_instance(g, s)
     graphptr g;
     substructure *s;
/* Returns TRUE if graph g does not already occur in the instances of
   substructure s.  If Overlap=TRUE, then the graph can overlap an
   instance of the substructure. */
{
  shortint i;
  graph_list_node_ptr gptr;

  if (! Overlap)
    for (i=0; i<g->nv; i++)
      if (vertex_in_sub(g->vertices[i]->number, s))
	return(FALSE);
  gptr = s->instances;
  while (gptr != NULL) {
    if (! distinct_graph(g, gptr->g, NULL))
      return(FALSE);
    gptr = gptr->next;
  }
  return(TRUE);
}


boolean distinct_graph(g1, g2, e)
     graphptr g1, g2;
     eptr e;
/* Returns TRUE if g1 is not equal to (g2 + e), where edge e
   can be NULL, but should not be in E(g2). */
{
  shortint i;
  if (e == NULL) {
    if (g1->ne != g2->ne)
      return(TRUE);
  } else if ((g2->ne != (g1->ne - 1)) || new_edge(e, g1))
    return(TRUE);
  for (i=0; i<g2->ne; i++)
    if (new_edge(g2->edges[i], g1)) return(TRUE);
  return(FALSE);
}


boolean unique_graph(g, e, glist)
     graphptr g;
     eptr e;
     graph_list_node_ptr glist;
/* Returns TRUE is graph (g + e) is not on the list of graphs glist. */
{
  while (glist != NULL) {
    if (! distinct_graph(glist->g, g, e)) return(FALSE);
    glist = glist->next;
  }
  return(TRUE);
}


boolean vertex_in_graph(num, g)
     shortint num;
     graphptr g;
/* Returns TRUE if vertex num is in graph g. */
{
  shortint i;
  for (i=0; i<g->nv; i++)
    if (num == g->vertices[i]->number)
      return(TRUE);
  return(FALSE);
}


boolean new_edge(e, g)
     eptr e;
     graphptr g;
/* Returns TRUE is edge e is not in graph g. */
{
  shortint i;
  for (i=0; i<g->ne; i++)
    if (g->edges[i] == e)
      return(FALSE);
  return(TRUE);
}

graph_list_node_ptr add_edge(insts)
     graph_list_node_ptr insts;
/* Returns a list of new instances created by extending the given instances
   by one edge in all possible ways. */
{
  graph_list_node_ptr inew, newinsts = NULL;
  graphptr oldg, newg;
  boolean infrom, into;
  shortint i, j, vnum;
  shortint addedge; /* 0 -> no edge to add, 1 -> add edge but no vertices,
                       2 -> add edge and vertex vnum */
  
  while (insts != NULL) {  /* Sequence through the list of instance graphs */
    oldg = insts->g;
    if (oldg->matchcost == 0.0) {
      for (i=0; i<Graph->ne; i++) {
	/* Find incident edge, if any */
	infrom = vertex_in_graph(Graph->edges[i]->from, oldg);
	into = vertex_in_graph(Graph->edges[i]->to, oldg);
	addedge = 0;
	if (infrom && (! into)) {
	  vnum = Graph->edges[i]->to;
	  addedge = 2;
	} else if ((! infrom) && into) {
	  vnum = Graph->edges[i]->from;
	  addedge = 2;
	} else if (infrom && into && (new_edge(Graph->edges[i], oldg)))
	  addedge = 1;
	/* If possible extension found, see if it would produce a unique
           graph.  If so, create new graph and add to list */
	if ((addedge > 0) && (unique_graph(oldg, Graph->edges[i], newinsts))) {
	  inew = (graph_list_node_ptr) malloc(sizeof(graph_list_node));
	  inew->g = (graphptr) malloc(sizeof(graph));
	  inew->g->refcount = 0;
	  inew->next = newinsts;
	  newinsts = inew;
	  newg = inew->g;
	  if (addedge == 2)  /* If adding edge and vertex */
	    newg->nv = oldg->nv + 1;
	  else newg->nv = oldg->nv;
	  newg->ne = oldg->ne + 1;
	  newg->matchcost = 0.0;
	  newg->vertices = (vptr *) malloc(newg->nv * sizeof(vptr));
	  newg->edges = (eptr *) malloc(newg->ne * sizeof(eptr));
	  for (j=0; j<oldg->nv; j++)
	    newg->vertices[j] = oldg->vertices[j];
	  for (j=0; j<oldg->ne; j++)
	    newg->edges[j] = oldg->edges[j];
	  /* add new edge and (possibly) vertex */
	  if (addedge == 2)
	    newg->vertices[newg->nv - 1] = find_vertex(vnum,Graph);
	  newg->edges[newg->ne - 1] = Graph->edges[i];
	}
      }
    }
    insts = insts->next;
  }
  return(newinsts);
}
