/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* fileio.c

Read in graph from a file and transform into SUBDUE representation.
Each line of a graph file defines either a vertex or an edge.
Vertices must be defined before they are used in edges.  Vertex lines
have the form 'v <#> <label>', where <#> is any positive integer, and
<label> is any string (no longer than TOKENLEN-1).  Edge lines have
one of three forms: 'e <label> <v1> <v2>', 'd <label> <v1> <v2>' or 'u
<label> <v1> <v2>', where <v1> and <v2> are the <#>'s for the
corresponding vertices, and <label> is any string (no longer than
TOKENLEN-1).  Edges beginning with 'e' are assumed directed unless
Directed=FALSE, in which case all 'e' edges become undirected.  Edges
beginning with 'd' are always directed, and edges beginning with 'u'
are always undirected.  Anything following a COMMENT character (see
below) on a line is ignored.

*/

#include "subdue.h"

#define COMMENT  '%' /* comment character in graph file */


void ignore_line(fp)
     FILE *fp;
{
  int c;
  while ((c = getc(fp)) != EOF && (c != '\n')) ;
}


void read_token(fp, s)
     FILE *fp;
     char *s;
/* Reads a token from stream fp into string s. */
{
  int c;
  shortint index = 0;
  c = getc(fp); /* skip over whitespace and comments */
  while (c == ' ' || c == '\n' || c == '\t' || c == COMMENT) {
    if (c == COMMENT)
      ignore_line(fp);
    c = getc(fp);
  }
  while (c != EOF && c != ' ' && c != '\n' && c != '\t' && c != COMMENT) {
    s[index++] = c;
    c = getc(fp);
  }
  if (c == COMMENT)
    ignore_line(fp);
  s[index] = '\0';
}


shortint read_number(fp)
     FILE *fp;
{
  char s[TOKENLEN];
  shortint i;

  read_token(fp, s);
  if (strlen(s) == 0) {
    printf("Unexpected end of file.\n");
    exit(1);
  }
  if (sscanf(s,"%hu",&i) != 1) {
    printf("Read error: %s should be an unsigned integer.\n",s);
    exit(1);
  }
  else return(i);
}


shortint read_vertex(fp, g, nv)
     FILE *fp;
     graphptr g;
     shortint nv;
{
  char tempstr[TOKENLEN];

  g->vertices[nv]->number = read_number(fp);
  read_token(fp, tempstr);
  g->vertices[nv]->label = enter_label(tempstr);
  return(nv + 1);
}


shortint read_edge(fp, g, directed, ne)
     FILE *fp;
     graphptr g;
     boolean directed;
     shortint ne;
{
  char tempstr[TOKENLEN];

  read_token(fp, tempstr);
  g->edges[ne]->label = enter_label(tempstr);
  g->edges[ne]->from = read_number(fp);
  g->edges[ne]->to = read_number(fp);
  g->edges[ne]->directed = directed;
  return(ne + 1);
}


graphptr read_graph(filename)
     char *filename;
{
  shortint i, nv = 0, ne = 0;
  char s[TOKENLEN];
  graphptr g = (graphptr) malloc(sizeof(graph));
  FILE *fp;
  
  NumLabels = 0;
  LabelList = NULL;
  count_vertices_and_edges(filename, g);
  if ((fp=fopen(filename,"r")) == NULL) {
    printf("Unable to open graph file %s.\n",filename);
    exit(1);
  }
  g->refcount = 1;
  g->vertices = (vptr *) malloc(g->nv * sizeof(vptr));
  g->edges    = (eptr *) malloc(g->ne * sizeof(eptr));
  for (i=0; i<g->nv; i++)
    g->vertices[i] = (vptr) malloc(sizeof(vertex));
  for (i=0; i<g->ne; i++)
    g->edges[i] = (eptr) malloc(sizeof(edge));
  g->matchcost = 0.0;
  read_token(fp, s);
  while (strlen(s) != 0) {
    switch (s[0]) {
    case 'v':
      nv = read_vertex(fp, g, nv);
      break;
    case 'e':
      ne = read_edge(fp, g, Directed, ne);
      break;
    case 'd':
      ne = read_edge(fp, g, TRUE, ne);
      break;
    case 'u':
      ne = read_edge(fp, g, FALSE, ne);
      break;
    default:
      printf("Unrecognized line in graph file.\n");
      exit(1);
    }
    read_token(fp, s);
  }
  fclose(fp);
  return(g);
}


void count_vertices_and_edges(filename, g)
     char *filename;
     graphptr g;
{
  char c = '\n';
  shortint nv = 0, ne = 0;
  FILE *fp;

  if ((fp = fopen(filename,"r")) == NULL) {
    printf("Unable to open %s.\n",filename);
    exit(1);
  }
  while (c != EOF) {
    while ((c != '\n') && (c != EOF))
      c = getc(fp);
    if (c != EOF) {
      c = getc(fp);
      if (c == 'v')
	nv++;
      if ((c == 'e') || (c == 'd') || (c == 'u'))
	ne++;
    }
  }
  g->nv = nv;
  g->ne = ne;
  fclose(fp);
}


shortint enter_label(str)
     char *str;
/* Returns the index of string str in the global LabelList.  If not found,
   creates a new entry for the string. */
{
  shortint i;

  for (i=0; i<NumLabels; i++)
    if (strcmp(LabelList[i], str) == 0)
      return(i);
  NumLabels++;
  if (LabelList == NULL)
    LabelList = (char **) malloc(NumLabels * sizeof(char *));
  else LabelList = (char **) realloc(LabelList, NumLabels * sizeof(char *));
  LabelList[NumLabels - 1] = (char *) malloc((strlen(str)+1) * sizeof(char));
  strcpy(LabelList[NumLabels - 1], str);
  return(NumLabels - 1);
}
