/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* fm.c

Fuzzy Graph Match

   Input:   Two (possibly similar) graph structures and a bound on the number
            of mappings considered (nodes in search space)
   Output:  The cost of the match from g1 to g2

   This version of the inexact graph match performs a branch-and-bound search
   through the space of possible mappings.  Nodes with the highest amount
   of connectivity are considered first.  The cost of a mapping is computed
   as the number of transformations (node addition/deletion, node label
   change, edge addition/deletion, edge label change or direction change)
   weighted by the user-defined cost of each transformation.

   This graph match can make use of a bound on the search time.  This bound
   is defined in the max_node function.  Once max_node number of nodes have
   been expanded in the search space, then only the best mapping is considered
   from that level in the search tree down to the leaf level, where a complete
   mapping is found.  If max_node is set to n!, then a complete search will be
   performed.  If max_node is set to anything less, then the search will use
   less time with a possible decrease in match cost accuracy.

*/

#include "subdue.h"

#define pick_node(x) x


double max_node(n)
   shortint n;
{
  double dn = (double) n;
  return(pow(dn,4.0));
}


double factorial(n)
   shortint n;
/* Using upper bound on Stirling's Approximation:
       n! <= sqrt(2*pi*n)*(n/e)^(n+1/12n) */
{
  double dn = (double) n;
  if (n == 0)
    return(1.0);
  else return(sqrt(dn*PI2)*pow((dn/exp(1.0)),(dn + (1.0/(12.0*dn)))));
}


boolean exact_gmatch (g1, g2)
     graphptr g1, g2;
/* Return TRUE if g1 and g2 match exactly (cost=0). */
{
  float match = 1.0;
  if ((g1->nv == g2->nv) && (g1->ne == g2->ne))
    match = fm(g1, g2, factorial(g1->nv));
  if (match == 0.0)
    return(TRUE);
  else return(FALSE);
}


float gmatch (g1, g2)
     graphptr g1, g2;
/* Return the cost of matching graphs g1 and g2.  The larger graph is
   passed first to fm. */
{
  if ((g1->nv + g1->ne) > (g2->nv + g2->ne))
    return(fm(g1,g2,max_node(g1->nv)));
  else return(fm(g2,g1,max_node(g2->nv)));
}


float fm(g1, g2, max_nodes)
     graphptr g1, g2;
     double max_nodes;
/* Returns the cost of changing graph g1 into graph g2, where
   size(g1) >= size(g2). */
{
  int i, j, node1, node2, temp, *ordered, *size, *used, child, change;
  int sorted = TRUE, n1num;
  double nodenum;
  float tempcost, cost;
  list head = NULL, tail = NULL, first = NULL, last = NULL;
  list node = NULL, one = NULL, next = NULL, hash[MAXHASH], best = NULL;
  pairptr pairs;

  change = 0;              /* change == 1 indicates switch to quick search */
  nodenum = 0.0;           /* Number of nodes expanded in the search tree */
  cost = -1.0;
  pairs = (pairptr) malloc(g1->nv * sizeof(pair));     /* node->node mapping */
  used = (int *) malloc(g1->nv * sizeof(int));       /* nodes already mapped */
  ordered = (int *) malloc(g1->nv * sizeof(int));   /* order by connectivity */
  size = (int *) malloc(g1->nv * sizeof(int));  /* connectivity of each node */
  for (i=0; i<g1->nv; i++) {
    ordered[i] = i;
    size[i] = 0;
    for (j=0; j<g1->ne; j++)
      if ((g1->edges[j]->from == i) || (g1->edges[j]->to == i))
	size[i]++;
  }
  for (i=0; i<g1->nv-1; i++) {       /* Sort the nodes by their connectivity */
    for (j=i+1; j<g1->nv; j++)
      if (size[i] > size[j]) {
	temp = ordered[i];
	ordered[i] = ordered[j];
	ordered[j] = temp;
	temp = size[i];
	size[i] = size[j];
	size[j] = temp;
	sorted = FALSE;
      }
    if (sorted == TRUE)
      break;
  }
  /* Unsearched nodes in the search tree are kept on doubly-linked list */
  head = (list) malloc(sizeof(key));                  /* global search queue */
  node = (list) malloc(sizeof(key));                      /* node = head->to */
  tail = (list) malloc(sizeof(key));
  for (i=0; i<MAXHASH; i++)     /* Initialize the pointers in the hash table */
    hash[i] = tail;
  head->depth = -1;                   /* Start at the top of the search tree */
  head->to = node;
  node->depth = 0;
  node->cost_now = 0.0;
  node->to = tail;
  node->from = head;
  tail->depth = -1;    /* Depth of -1 signals the beginning / end of a queue */
  tail->from = head;
  first = (list) malloc(sizeof(key));          /* Local (child) search queue */
  first->depth = -1;
  last = (list) malloc(sizeof(key));
  last->depth = -1;
  while (node->depth >= 0) {              /* While search queue is not empty */
    nodenum++;
    /* Search time has expired, quickly find rest of mapping */
    if ((nodenum > max_nodes) && (change == 0)) {
      first->to = node;                /* Move global queue to local queue */
      node->from = first;
      tail->from->to = last;
      last->from = tail->from;
      head->to = tail;
      tail->from = head;
      i = 0;
      best = last;
      while (node->depth > -1) { /* Move down each level to bottom of tree */
	next = node->to;
	if (i < pick_node(g1->nv)) {    /* pick_node = number levels left */
	  if ((best->depth < 0) ||
	      (floor(node->cost_now) != floor(best->cost_now))) {
	    i++;
	    if (best->depth > -1) {
	      tail->from->to = best;
	      best->from = tail->from;
	      best->to = tail;
	      tail->from = best;
	    }
	    best = (list) malloc(sizeof(key));
	    best->pairing = (pairptr) malloc(g1->nv * sizeof(pair));
	    best->depth = node->depth;
	    best->cost_now = node->cost_now;
	    for (j=0; j<node->depth; j++) {
	      best->pairing[j].n1 = node->pairing[j].n1;
	      best->pairing[j].n2 = node->pairing[j].n2;
	    }
	  }
	  else if (node->depth >= best->depth) {
	    best->depth = node->depth;
	    best->cost_now = node->cost_now;
	    for (j=0; j<node->depth; j++) {
	      best->pairing[j].n1 = node->pairing[j].n1;
	      best->pairing[j].n2 = node->pairing[j].n2;
	    }
	  }
	}
	free(node->pairing);
	free(node);
	node = next;
      }
      tail->from->to = best;
      best->from = tail->from;
      best->to = tail;
      tail->from = best;
      change = 1;                /* Flag new type of search - quick search */
      node = head->to;
    }
    if (node->depth == g1->nv) { /* Hit the bottom level of the search tree */
      if ((node->cost_now < cost) || (cost < 0.0)) {
	cost = node->cost_now;
	for (i=0; i<g1->nv; i++) {
	  pairs[i].n1 = node->pairing[i].n1;
	  pairs[i].n2 = node->pairing[i].n2;
	}
      }
      if (change == 0) {
	/* Delete current node from search queue */
	while (node->depth > -1) {
	  next = node->to;
	  free(node->pairing);
	  free(node);
	  node = next;
	}
	head->to = tail;
	tail->from = head;
      }
      else {
	head->to = node->to;
	node->to->from = head;
	free(node->pairing);
	free(node);
      }
    }
    else {
      node1 = ordered[node->depth];
      n1num = g1->vertices[node1]->number;
      for (i=0; i<g1->nv; i++)
	used[i] = 0;
      j = g2->nv;
      for (i=0; i<node->depth; i++)
	if (node->pairing[i].n2 != DELETE)
	  used[node->pairing[i].n2] = 1;
	else {
	  used[j] = 1;
	  j++;
	}
      first->to = last;                   /* Initialize local search queue */
      last->from = first;
      for (child=0; child<g1->nv; child++) {
	if (used[child] != 1) {                  /* This is a new mapping */
	  node2 = child;
	  if (node2 >= g2->nv)     /* There are more nodes in g2 than g1 */
	    node2 = DELETE;
	  /* Increase overall cost by cost of this partial match */
	  tempcost = 0.0;
	  if (node2 == DELETE) {
	    tempcost += DELNODE;
	    for (i=0; i<g1->ne; i++)
	      if ((g1->edges[i]->from == n1num) ||
		  (g1->edges[i]->to == n1num))
		tempcost += DELEDGEP;
	  }
	  else {
	    if (g1->vertices[node1]->label != g2->vertices[node2]->label)
	      tempcost += SUBNODE;
	    tempcost += deleted_edges(g1, g2, node->pairing, n1num,
			      g2->vertices[node2]->number, node->depth);
	    tempcost += inserted_edges(g1, g2, node->pairing, n1num,
			      g2->vertices[node2]->number, node->depth);
	  }
	  /* Expand next node (partial match) in search tree */
	  one = (list) malloc(sizeof(key));
	  one->depth = node->depth + 1;
	  one->cost_now = node->cost_now + tempcost;
	  one->pairing = (pairptr) malloc(one->depth * sizeof(pair));
	  if (one->pairing == NULL) {
	    printf("RAN OUT OF MEMORY\n");
	    exit(1);
	  }
	  for (i=0; i<node->depth; i++) {
	    one->pairing[i].n1 = node->pairing[i].n1;
	    one->pairing[i].n2 = node->pairing[i].n2;
	  }
	  one->pairing[node->depth].n1 = node1;
	  one->pairing[node->depth].n2 = node2;
	  
	  /* Insert new node in local search queue, ordering queue by
	     cost-so-far */
	  next = first->to;
	  while ((one->cost_now > next->cost_now) && (next->depth > -1))
	    next = next->to;
	  next->from->to = one;
	  one->from = next->from;
	  next->from = one;
	  one->to = next;
	}
      }
      if (change == 0) {
	head->to = node->to;              /* Pop node off of global queue */
	node->to->from = head;
	
	/* Set a new pointer in the hash table to the next significant
	   (determined by SCALE) cost-so-far in the current search queue */
	for (i=0; (i<=floor(node->cost_now*SCALE)) && (i<MAXHASH); i++)
	  hash[i] = node->to;
	free(node);
	one = first->to;
	while (one->depth > -1) {
	  next = one->to;
	  i = (int) floor((double) one->cost_now*SCALE);
	  if (i >= MAXHASH)
	    i = MAXHASH - 1;
	  hash[i]->from->to = one;
	  one->from = hash[i]->from;
	  one->to = hash[i];
	  hash[i]->from = one;
	  if (one->from->depth > -1)
	    j = (int) floor((double) one->from->cost_now*SCALE) + 1;
	  else j = 0;
	  for (; j<=i; j++)
	    hash[j] = one;
	  one = next;
	}
      }
      else {
	one = first->to->to;
	first->to->to = head->to->to;
	head->to->to->from = first->to;
	head->to = first->to;
	first->to->from = head;
	free(node->pairing);
	free(node);
	while (one->depth > -1) {
	  next = one->to;
	  free(one->pairing);
	  free(one);
	  one = next;
	}
      }
    }
    node = head->to;
  }
  return(cost);
}


/*                         DELETED_EDGES

     Input:  g1 and g2:  two graphs being matched
	     pairs:      array of current node->node matches
	     n1 and n2:  current node->node match being considered
	     size:       number of elements in the PAIRS array

     Output: Cost computed for this partial match based on edges that must
	     be deleted from the first partial graph or edge labels that must
	     be changed in the first partial graph to make it isomorphic to
	     the second partial graph.
*/

float deleted_edges(g1, g2, pairs, n1, n2, size)
     graphptr g1, g2;
     pairptr pairs;
     int n1, n2;
     int size;
{
  int i, j, found, m;
  float cost = 0.0;

  /* Scan through edges, finding those connected to n1 */
  for (i=0; i<g1->ne; i++) {
    if (g1->edges[i]->from == n1) {
      m = get_match(g1->edges[i]->to, size, pairs, g1, g2, FORWARD);
      if (m != -2) {
	found = FALSE;
	for (j=0; j<g2->ne; j++)    /* Does g2 have a corresponding edge? */
	  if (((g1->edges[j]->directed == TRUE) &&
	       ((g2->edges[j]->from == n2) && (g2->edges[j]->to == m))) ||
	      ((g1->edges[j]->directed == FALSE) &&
	       (((g2->edges[j]->from == n2) && (g2->edges[j]->to == m)) ||
		((g2->edges[j]->from == m) && (g2->edges[j]->to == n2))))) {
	    found = TRUE;                /* g2 has a corresponding edge */
	    /* The edge labels are different */
	    if (g1->edges[i]->label != g2->edges[j]->label)
	      cost += SUBEDGE;
	    /* One edge is directed, the other is undirected */
	    if (g1->edges[i]->directed != g2->edges[j]->directed)
	      cost += SUBEDGEDIR;
	    break;
	  }
	if (found == FALSE)   /* g2 does not have a corresponding edge */
	  cost += DELEDGE;
      }
    }
    if (g1->edges[i]->to == n1) {
      m = get_match(g1->edges[i]->from, size, pairs, g1, g2, FORWARD);
      if (m != -2) {
	found = FALSE;
	for (j=0; j<g2->ne; j++)    /* Does g2 have a corresponding edge? */
	  if (((g1->edges[j]->directed == TRUE) &&
	       ((g2->edges[j]->to == n2) && (g2->edges[j]->from == m))) ||
	      ((g1->edges[j]->directed == FALSE) &&
	       (((g2->edges[j]->to == n2) && (g2->edges[j]->from == m)) ||
		((g2->edges[j]->to == m) && (g2->edges[j]->from == n2))))) {
	    found = TRUE;                /* g2 has a corresponding edge */
	    /* The edge labels are different */
	    if (g1->edges[i]->label != g2->edges[j]->label)
	      cost += SUBEDGE;
	    /* One edge is directed, the other is undirected */
	    if (g1->edges[i]->directed != g2->edges[j]->directed)
	      cost += SUBEDGEDIR;
	    break;
	  }
      }
      if (found == FALSE)       /* g2 does not have a corresponding edge */
	cost += DELEDGE;
    }
  }
  return(cost);
}


/*                        INSERTED_EDGES

     Input:  g1 and g2:  two graphs being matched
	     pairs:      array of current node->node matches
	     n1 and n2:  current node->node match being considered
	     size:       number of elements in the PAIRS array

     Output: Cost computed for this partial match based on edges that must
	     be inserted to the first partial graph to make it isomorphic
	     to the second partial graph.
*/

float inserted_edges(g1, g2, pairs, n1, n2, size)
     graphptr g1, g2;
     pairptr pairs;
     int n1, n2;
     int size;
{
  int i, j, found, m;
  float cost = 0.0;
  
  /* Scan through edges, finding those connected to n2 */
  for (i=0; i<g2->ne; i++) {
    if (g2->edges[i]->from == n2) {
      m = get_match(g2->edges[i]->to, size, pairs, g1, g2, BACKWARD);
      if (m != -2) {
	found = FALSE;
	for (j=0; j<g1->ne; j++)    /* Does g1 have a corresponding edge? */
	  if (((g1->edges[j]->directed == TRUE) &&
	       ((g1->edges[j]->from == n1) && (g1->edges[j]->to == m))) ||
	      ((g1->edges[j]->directed == FALSE) &&
	       (((g1->edges[j]->from == n1) && (g1->edges[j]->to == m)) ||
		((g1->edges[j]->from == m) && (g1->edges[j]->to == n1)))))
	    {
	      found = TRUE;                /* g1 has a corresponding edge */
	      break;
	    }
	if (found == FALSE)      /* g1 does not have a corresponding edge */
	  cost += INSEDGE;
      }
    }
    if (g2->edges[i]->to == n2) {
      m = get_match(g2->edges[i]->from, size, pairs, g1, g2, BACKWARD);
      if (m != -2) {
	found = FALSE;
	for (j=0; j<g1->ne; j++)    /* Does g1 have a corresponding edge? */
	  if (((g1->edges[j]->directed == TRUE) &&
	       ((g1->edges[j]->to == n1) && (g1->edges[j]->from == m))) ||
	      ((g1->edges[j]->directed == FALSE) &&
	       (((g1->edges[j]->to == n1) && (g1->edges[j]->from == m)) ||
		((g1->edges[j]->to == m) && (g1->edges[j]->from == n1)))))
	    {
	      found = TRUE;                /* g1 has a corresponding edge */
	      break;
	    }
	if (found == FALSE)      /* g1 does not have a corresponding edge */
	  cost += INSEDGE;
      }
    }
  }
  return(cost);
}


int get_match(node, size, pairs, g1, g2, dir)
     int node, size;
     pairptr pairs;
     graphptr g1, g2;
     int dir;
/* Given a node from one graph, find its corresponding node in a second graph
   as specified by the list of matches in the PAIRS array.
   The variable DIR specifies which half of the match to consider. */
{
  int i=0;

  while ((i < size) && (pairs[i].n1 != -1)) {
    if (dir == FORWARD) {
      if (g1->vertices[pairs[i].n1]->number == node) {
	if (pairs[i].n2 == -2)
	  return(-2);
	else return(g2->vertices[pairs[i].n2]->number);
      }
    }
    else {
      if (pairs[i].n2 != -2) {
	if (g2->vertices[pairs[i].n2]->number == node) {
	  if (pairs[i].n1 == -1)
	    return(-2);
	  else return(g1->vertices[pairs[i].n1]->number);
	}
      }
    }
    i++;
  }
  return(-2);
}
