/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* subdue.c

   subdue [-limit #] [-iterations #] [-threshold #] [-trace #]
          [-n nsubs] [-noprune] [-beam #] [-overlap] [-undirect]
          [-con connectivity] [-com compactness] [-cov coverage]
          graphfile

   [-limit #] The number of different substructures to consider.
          Default is zero, which implies a limit of 0.5*E(G).

   [-iterations #] The number of iterations made over the input graph in
          which the best substructure from the previous iteration is used
          to compress the graph for use in the next iteration.  Default
          is one, which implies only one pass, no compression.

   [-threshold #] The fraction of the size of an instance by which the
          instance can be different (according to the distortion
          costs) from the substructure definition.  I.e., the graphs
          match if matchcost(sub,inst) <= size(inst)*threshold.
          Default is 0.0, which implies graphs must match exactly.

   [-trace #] Sets the level of output produced by Subdue.  Zero is the
          default, which produces minimal output; namely, the best nsubs
          discovered substructures.  Trace=1 adds all substructures as
          they are considered by the discovery procedure.  Trace=2 adds
          the instances to printed substructures.  Trace=3 adds true
          tracing of some procedures (used mostly for debugging).

   [-nsubs #] Retains no more than the top #>0 substructures, which are
          returned.  Only the best substructure is used for replacement.
          Default is 3.

   [-noprune] Subdue will discard substructures with less value than their
          parent substructure.  Specifying this argument will prevent this.

   [-beam #] Limits the number of substructures waiting for consideration
          during the discovery process.  Default is 4.

   [-overlap] Subdue normally will not allow overlap among the instances
          of a substructure.  Specifying this argument will allow overlap.
          Note that when allowing overlap, compressing the instances of
          a substructure requires extra information indicating where the
          overlap occurs.  This tends to hinder compression due to the
          additional bits needed to describe the overlap.  Also, if overlap
          is allowed, substructures will have more instances, which tends
          to inflate their heuristic value.

   [-undirect] Subdue assumes that edges in the input graph file defined
          using 'e' (and edges created to describe overlapping graphs) are
          directed edges.  Specifying this argument makes these edges
          undirected.  Note that graph file edges defined with 'u' are
          always undirected, and edges defined with 'd' are always directed.

   [-con connectivity] The exponent to the connectivity rule value of a
          substructure and its instances.  The value measures the average
          number of connections between the instances and the rest of the
          input graph.  Default is 0.0, which implies this rule has no
          effect on the evaluation of a substructure.  A positive value
          favors substructures with low connectivity, and a negative
          value favors substructures with high connectivity.

   [-com compactness] The exponent to the compactness rule value of a
          substructure and its instances.  The value measures the average
          compactness of the instances.  Default is 0.0, which implies this
          rule has no effect on the evaluation of a substructure.  A positive
          value favors compact (closed) substructures, and a negative value
          favors less dense substructures.

   [-cov coverage] The exponent to the coverage rule value of a substructure
          and its instances.  The value measures the fraction of the input
          graph covered by the instances.  Default is 0.0, which implies this
          rule has no effect on the evaluation of a substructure.  A positive
          value favors substructures with high coverage, and a negative value
          favors substructures with low coverage.

*/

#include "subdue.h"


void usage()
{
  printf("usage: subdue [-limit #] [-iterations #] [-threshold #] [-trace #]\n");
  printf("              [-nsubs #] [-noprune] [-beam #] [-overlap] [-undirect]\n");
  printf("              [-con connectivity] [-com compactness] [-cov coverage]\n");
  printf("              graphfile\n");
  exit(1);
}


void print_parameters(limit,iterations,bestmdl,g)
     shortint limit,iterations;
     float bestmdl;
     graphptr g;
{
  printf("Subdue Parameters:\n");
  printf("  Limit = %u\n",limit);
  printf("  Iterations = %u\n",iterations);
  printf("  Threshold = %f\n",Threshold);
  printf("  Trace = %u\n",Trace);
  printf("  NumSubs = %u\n",NumSubs);
  printf("  Prune = ");
  if (Prune) printf("TRUE\n"); else printf("FALSE\n");
  printf("  Beam = %u\n",Beam);
  printf("  Overlap = ");
  if (Overlap) printf("TRUE\n"); else printf("FALSE\n");
  printf("  Directed = ");
  if (Directed) printf("TRUE\n"); else printf("FALSE\n");
  printf("  Connectivity = %f\n",Connectivity);
  printf("  Compactness = %f\n",Compactness);
  printf("  Coverage = %f\n",Coverage);
  printf("  Input Graph: %d vertices, %d edges, mdl = %f\n",
	 g->nv,g->ne,bestmdl);
}


void main(argc, argv)
   int argc;
   char *argv[];
{
  graphptr g, ng;
  sub_list_node_ptr subs;
  int tmp;
  shortint i = 0, iterations = 1, limit = 0;
  float bestmdl, mdlval;

  Trace = 0;
  Beam = 4;
  Directed = TRUE;
  Overlap = FALSE;
  Prune = TRUE;
  Threshold = 0.0;
  NumSubs = 3;
  Connectivity = 0.0;
  Compactness = 0.0;
  Coverage = 0.0;
  while ((--argc > 0) && ((*++argv)[0] == '-')) {
    if (strcmp(*argv,"-limit") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%d",&tmp) == 0) {
	printf("subdue: bad argument %s to -limit option\n",*argv);
	argc = 0;
      }
      limit = (shortint) tmp;
    } else if (strcmp(*argv,"-iterations") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%d",&tmp) == 0) {
	printf("subdue: bad argument %s to -iterations option\n",*argv);
	argc = 0;
      }
      iterations = (shortint) tmp;
    } else if (strcmp(*argv,"-nsubs") == 0) {
      argv++; argc--;
      if ((sscanf(*argv,"%d",&tmp) == 0) || (tmp <= 0)) {
	printf("subdue: bad argument %s to -nsubs option\n",*argv);
	argc = 0;
      }
      NumSubs = (shortint) tmp;
    } else if (strcmp(*argv,"-threshold") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%f",&Threshold) == 0) {
	printf("subdue: bad argument %s to -threshold option\n",*argv);
	argc = 0;
      }
    } else if (strcmp(*argv,"-trace") == 0) {
      argv++; argc--;
      if ((sscanf(*argv,"%d",&tmp) == 0) || (tmp <= 0)) {
	printf("subdue: bad argument %s to -trace option\n",*argv);
	argc = 0;
      }
      Trace = (shortint) tmp;
    } else if (strcmp(*argv,"-beam") == 0) {
      argv++; argc--;
      if ((sscanf(*argv,"%d",&tmp) == 0) || (tmp <= 0)) {
	printf("subdue: bad argument %s to -beam option\n",*argv);
	argc = 0;
      }
      Beam = (shortint) tmp;
    } else if (strcmp(*argv,"-noprune") == 0) {
      Prune = FALSE;
    } else if (strcmp(*argv,"-overlap") == 0) {
      Overlap = TRUE;
    } else if (strcmp(*argv,"-undirect") == 0) {
      Directed = FALSE;
    } else if (strcmp(*argv,"-con") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%f",&Connectivity) == 0) {
	printf("subdue: bad argument %s to -con option\n",*argv);
	argc = 0;
      }
    } else if (strcmp(*argv,"-com") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%f",&Compactness) == 0) {
	printf("subdue: bad argument %s to -com option\n",*argv);
	argc = 0;
      }
    } else if (strcmp(*argv,"-cov") == 0) {
      argv++; argc--;
      if (sscanf(*argv,"%f",&Coverage) == 0) {
	printf("subdue: bad argument %s to -cov option\n",*argv);
	argc = 0;
      }
    } else {
      printf("subdue: illegal option %s\n",*argv);
      argc = 0;
      break;
    }
  }
  if (argc != 1) usage();
  g = read_graph(*argv);
  bestmdl = (float) mdl(g); 
  print_parameters(limit,iterations,bestmdl,g);
  while (i<iterations) {
    i++;
    printf("\n----- Iteration %u -----\n",i);
    subs = subdue(g,limit);
    if (subs == NULL) {
      printf("\nNo substructures to replace.\n");
      i = iterations;  /* no reason to continue */
    } else {
      printf("\nFound the following substructures:\n\n");
      print_sub_list(subs);
      if (i<iterations) { /* another iteration to go, try to compress */
	ng = compress(g, subs->s, i);
	mdlval = (float) mdl(ng);
	mdlval += (float) mdl(subs->s->definition);
	if (mdlval < bestmdl) {
	  bestmdl = mdlval;
	  printf("Compressed graph using substructure:\n\n");
	  print_substructure(subs->s);
	  clean_up(subs,g);
	  g = ng;
	} else {
	  printf("\nUnable to compress graph further.\n");
	  i = iterations;  /* no reason to continue */
	}
      }
    }
  } /* end of while loop */
}


void clean_up(subs,g)
     sub_list_node_ptr subs;
     graphptr g;
/* Free all memory used by graph g and the graphs in subs. */
{
  graph_list_node_ptr glist, glnptr1, glnptr2;
  graphptr g1, g2;
  shortint i, j;

  glist = (graph_list_node_ptr) malloc(sizeof(graph_list_node));
  glist->g = g;
  glist->next = consolidate(subs);
  /* Free graphs in glist one by one, ensuring that shared vertex and
     edge objects are freed only once. */
  while (glist != NULL) {
    glnptr1 = glist;
    glist = glist->next;
    g1 = glnptr1->g;
    for (i=0; i<g1->nv; i++)  /* free vertices */
      if (g1->vertices[i] != NULL) {
	glnptr2 = glist;
	while (glnptr2 != NULL) {
	  g2 = glnptr2->g;
	  for (j=0; j<g2->nv; j++)
	    if (g1->vertices[i] == g2->vertices[j])
	      g2->vertices[j] = NULL;
	  glnptr2 = glnptr2->next;
	}
	free(g1->vertices[i]);
      }
    free(g1->vertices);
    for (i=0; i<g1->ne; i++)  /* free edges */
      if (g1->edges[i] != NULL) {
	glnptr2 = glist;
	while (glnptr2 != NULL) {
	  g2 = glnptr2->g;
	  for (j=0; j<g2->ne; j++)
	    if (g1->edges[i] == g2->edges[j])
	      g2->edges[j] = NULL;
	  glnptr2 = glnptr2->next;
	}
	free(g1->edges[i]);
      }
    free(g1->edges);
    free(g1);
    free(glnptr1);
  }
}


graph_list_node_ptr consolidate(subs)
  sub_list_node_ptr subs;
/* Put all graph definitions and instances in subs onto one list. */
{
  graph_list_node_ptr glnptr, glnptr1, glist = NULL;
  sub_list_node_ptr sptr, sptr1;
  substructure *s;

  sptr = subs;
  while (sptr != NULL) {
    s = sptr->s;
    if (s->definition != NULL) {
      glnptr = (graph_list_node_ptr) malloc(sizeof(graph_list_node));
      glnptr->g = s->definition;
      glnptr->next = glist;
      glist = glnptr;
    }
    glnptr = s->instances;
    while (glnptr != NULL) {
      glnptr1 = glnptr;
      glnptr = glnptr->next;
      glnptr1->next = glist;
      glist = glnptr1;
    }
    free(s);
    sptr1 = sptr;
    sptr = sptr->next;
    free(sptr1);
  }
  return glist;
}
