/*   THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
     AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
     (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
     WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
     WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
     ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
     RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
     JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
     DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
     generality of the foregoing, neither the author, nor JAIR, nor JAIR's
     publishers and distributors, warrant that the Source Code will be 
     error-free, will operate without interruption, or will meet the needs 
     of the user.
*/


/* subdue.h

  SUBDUE data structures.
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#define FALSE      0
#define TRUE       1
#define TOKENLEN   80           /* max label length = TOKENLEN-1 */
#define PI2        6.283185307  /* 2*pi, used in factorial computations */


/* The following definitions assign costs to the various transformations
   for changing one graph into another.  Eventually, these may be made
   into Subdue arguments to facilitate user preferences.  Note that our
   graph match ensures that we are always transforming a larger graph
   into a smaller graph.  Therefore, transformations involving the addition
   of a vertex (INSNODE and INSEDGEP) are not used, but are included for
   completeness. */

#define INSNODE    1.0  /* Cost of an insert-node graph transformation */
#define DELNODE    1.0  /* Cost of a delete-node graph transformation */
#define SUBNODE    1.0  /* Cost of a substitute-node-label transformation */
#define INSEDGE    1.0  /* Cost of an insert-edge graph transformation */
#define INSEDGEP   1.0  /* Cost of an insert-edge-with-node transformation */
#define DELEDGE    1.0  /* Cost of a delete-edge graph transformation */
#define DELEDGEP   1.0  /* Cost of a delete-edge-with-node transformation */
#define SUBEDGE    1.0  /* Cost of a substitute-edge-label transformation */
#define SUBEDGEDIR 1.0  /* Cost of a directed-undirected edge transformation */

/* The remaining definitions pertain to the fuzzy graph match (fm)
   procedure */

#define DELETE     -2   /* Value of an empty node (used in the pairings),
			   this represents a node that must be deleted */

#define FORWARD    1    /* Used in get_match, FORWARD and BACKWARD represent */
#define BACKWARD   2    /*  which half of the pairings to consider */

#define MAXHASH    10   /* Maximum size of hash table used in graph match */
#define SCALE       1   /* Scale used to set entries in hash table */


/***** Type Definitions *****/

typedef unsigned short shortint;
typedef unsigned char boolean;

typedef struct vobj {              /* Vertex */
  shortint number;
  shortint label;
  } vertex, *vptr;

typedef struct eobj {              /* Edge */
  shortint label;
  shortint from, to;
  boolean directed;
  } edge, *eptr;

typedef struct gobj {              /* Graph */
  shortint nv, ne;     /* number of vertices and edges */
  vptr *vertices;      /* pointer to array of pointers to vertices */
  eptr *edges;         /* pointer to array of pointers to edges */
  float matchcost;     /* if instance */
  shortint refcount;   /* number of times this graph is referenced */
} graph, *graphptr;

typedef struct glnode {
  graphptr g;
  struct glnode *next;
} graph_list_node, *graph_list_node_ptr;

typedef struct subobj {            /* Substructure */
  graphptr definition;
  graph_list_node_ptr instances;
  float value;         /* heuristic value of substructure */
  } substructure;

typedef struct slnode {            /* List of substructures */
  substructure *s;
  struct slnode *next;
  } substructure_list_node, *sub_list_node_ptr;

typedef struct pairobj {           /* List of vertex mappings */
  int n1, n2;
} pair, *pairptr;

/* Structure used for entires in the graph match search queue and hash table */
typedef struct keyobj {
   int depth;
   float cost_now;
   pairptr pairing;
   struct keyobj *from, *to;
} key, *list;


/***** Function Prototypes *****/

/* compress.c */

graphptr compress();
void connect();
boolean edge_in_sub(), share_vertex(), share_edge(), vertex_in_sub();
shortint newvertnum(), find_overlaps(), add_share();

/* discover.c */

sub_list_node_ptr discover_subs(), insert_sub(), best_n_subs();
boolean member_sub();

/* eval.c */

void evaluate_sub();
boolean member_vertex(), member_edge();
float instance_weight(), unique_structure();
float rules(), connectivity(), compactness(), coverage();

/* extend.c */

sub_list_node_ptr extend_sub();
void add_instances();
boolean distinct_graph(), unique_graph(), unique_instance();
boolean vertex_in_graph(), new_edge(); 
graph_list_node_ptr add_edge();

/* fileio.c */

void ignore_line(), read_token();
void count_vertices_and_edges();
shortint read_number(), read_vertex(), read_edge(), enter_label();
graphptr read_graph();

/* fm.c */

float gmatch(), fm(), deleted_edges(), inserted_edges();
int get_match();
double max_node(), factorial();
boolean exact_gmatch();

/* main.c */

void usage(), print_parameters(), clean_up();
graph_list_node_ptr consolidate();

/* mdl.c */

double mdl(), log2comb(), log2nfact(), log2();
shortint max_edge(), row_edge();

/* subdue.c */

boolean label_member();
sub_list_node_ptr subdue(), initial_subs();
substructure *initial_substructure();
graph_list_node_ptr get_instances();

/* util.c */

graphptr build_graph();
void free_substructure();
shortint size();
vptr find_vertex();
void print_vertices(), print_edges(), print_graph();
void print_substructure(), print_sub_list();


/***** Global Variables *****/

graphptr Graph;  /* Graph currently being search by Subdue */
shortint Size;   /* Number of vertices and edges in Graph */

boolean Directed; /* If TRUE, edges defined with 'e' are directed, else not */
boolean Overlap;  /* If TRUE, instances of a substructure may overlap. */
boolean Prune;    /* If TRUE, discard substructures with less value than
                     their parent substructure. */
shortint Beam;    /* Limit on substructures waiting for consideration */
shortint NumSubs; /* Number of best substructures retained by Subdue */
shortint Limit;   /* Number of substructures to generate */
float Threshold;  /* [0,1] matchcost must be below Threshold*Size(subgraph)
                     to be considered an instance */
float Connectivity; /* Exponents to heuristic rule values during evaluation */
float Compactness;
float Coverage;

shortint Trace;   /* Indicates amount of output, zero outputs least */
char **LabelList; /* Labels used in graph and substructures */
int NumLabels;    /* Number of unique labels in the orginal graph and those
                     created to record overlap during compression. */
